
/**************************************************************************
 *                                                                        *
 * This code is developed by Adam Li.  This software is an                *
 * implementation of a part of one or more MPEG-4 Video tools as          *
 * specified in ISO/IEC 14496-2 standard.  Those intending to use this    *
 * software module in hardware or software products are advised that its  *
 * use may infringe existing patents or copyrights, and any such use      *
 * would be at such party's own risk.  The original developer of this     *
 * software module and his/her company, and subsequent editors and their  *
 * companies (including Project Mayo), will have no liability for use of  *
 * this software or modifications or derivatives thereof.                 *
 *                                                                        *
 * Project Mayo gives users of the Codec a license to this software       *
 * module or modifications thereof for use in hardware or software        *
 * products claiming conformance to the MPEG-4 Video Standard as          *
 * described in the Open DivX license.                                    *
 *                                                                        *
 * The complete Open DivX license can be found at                         *
 * http://www.projectmayo.com/opendivx/license.php .                      *
 *                                                                        *
 **************************************************************************/

/**************************************************************************
 *
 *  mom_bitstream.c
 *
 *  Copyright (C) 2001  Project Mayo
 *
 *  Adam Li
 *
 *  DivX Advance Research Center <darc@projectmayo.com>
 *
 **************************************************************************/

/* This file contains some utility functions to write to bitstreams.      */
/* Some codes of this project come from MoMuSys MPEG-4 implementation.    */
/* Please see seperate acknowledgement file for a list of contributors.   */

#include <stdio.h>
#include "mom_bitstream.h"
#include "vm_enc_defs.h"


Image *
BitstreamInit()
{
	Image *bitstream;

	/* avoids problems with allocating zero size */

	bitstream = AllocImage(BITSTREAM_ALLOC_SIZE, 1, SHORT_TYPE);

	/* however, allocated size is not equal to x to avoid too many reallocations */
	/* there is no access function to store x, this is another reason that
	   macros would make more sense than functions for this purpose */

	bitstream->x = 0;
	return bitstream;
}


Void
BitstreamFree(Image *bitstream)
{
	FreeImage(bitstream);
}


Void
BitstreamPutBit(Image *bitstream, SInt bit)
{
	UInt size_x;

	size_x = GetImageSizeX(bitstream);
	if (size_x > 0 && size_x % BITSTREAM_ALLOC_SIZE == 0)
	{
		ReallocImage(bitstream, size_x+BITSTREAM_ALLOC_SIZE, 1);
		bitstream->x = size_x;
	}

	/* it was a good idea to use long instead of short for the sizes */

	bitstream->f[bitstream->x++] = bit;
}


Void
BitstreamPutBits(Image *bitstream, UInt bits, UInt nbits)
{
	UInt mask;

	/* it should be 1 <= nbits <= 32 */

	if (nbits > 0)
	{
		for (mask = 1L << (nbits-1); mask; mask >>= 1)
			BitstreamPutBit(bitstream, (SInt)((bits & mask) == mask));
	}
}


UInt
BitstreamPut (Image *b_stream, Int index1, Int index2)
{
	UInt i;
	SInt *stream = (SInt *)GetImageData (b_stream);
	UInt size_stream = GetImageSizeX (b_stream);

	for (i = 0; i < size_stream; i++)
		PutBits (1, *(stream++), index1,index2);

	return(size_stream);
}

/* the putbits functions */

/* static data for pointers and counters */
static UChar outbfr[MAX_NUM_VOS][MAX_NUM_VOLS];
static Int outcnt[MAX_NUM_VOS][MAX_NUM_VOLS];
static Int bytecnt[MAX_NUM_VOS][MAX_NUM_VOLS];

Void
InitBits(Int index1,Int index2)
{
	outcnt[index1][index2] = 8;
	bytecnt[index1][index2] = 0;
}


void PutByteStream(unsigned char byte);

Void
PutBits (Int n, Int val, Int index1, Int index2)
{
	Int i;
	U_Int mask;

	mask = 1 << (n-1);							  /* selects first (leftmost) bit */

	for (i=0; i<n; i++)
	{
		outbfr[index1][index2] <<= 1;

		if (val & mask)
			outbfr[index1][index2]|= 1;

		mask >>= 1;								  /* select next bit */
		outcnt[index1][index2]--;

		if (outcnt[index1][index2]==0)			  /* 8 bit buffer full */
		{
			// putc(outbfr[index1][index2],bitstream[index1][index2]);
			// hacked to catch the output stream <adamli>
			PutByteStream(outbfr[index1][index2]);
			outcnt[index1][index2] = 8;
			bytecnt[index1][index2]++;
		}
	}
}


Int
CloseBits (Int index1,Int index2)
{
	Int i;
	Int count, ret_value;

	if (outcnt[index1][index2]!=8)
	{
		count = outcnt[index1][index2];
		for (i = 0; i < count; i++)
			PutBits (1, 1, index1, index2);
		ret_value = outcnt[index1][index2];
		/* outcnt is reset in call to putbits () */
		return ret_value;
	}
	else
		return 0;
}


Int
NextStartCode(Int index1, Int index2)
{

	Int count;
	Int   i;

	count = outcnt[index1][index2] - 1;
	i = 0;

	PutBits(1,0,index1,index2);

	while (i < count)
	{
		PutBits(1,1,index1,index2);
		i++;
	}

	return (count+1);
}


Int
NextResyncMarker(Int index1, Int index2)
{

	Int count;
	Int   i;
	Int ret_value;

	ret_value = outcnt[index1][index2];

	count = outcnt[index1][index2] - 1;
	i = 0;

	PutBits(1,0,index1,index2);

	while (i < count)
	{
		PutBits(1,1,index1,index2);
		i++;
	}

	return ret_value;

}
